import subprocess
import os
import coast.coast_var as coast_var
import requests
import time
import json
from urllib import request
from shutil import unpack_archive
import hashlib


class BlastBatch:
    def __init__(self,
                 query,
                 db,
                 out="blast_results.tab",
                 taxonlist="",
                 taxidlist="",
                 task="",
                 matrix="",
                 evalue="",
                 gapextend="",
                 gapopen=""):
        """
        Formulate BLAST+ command line query.

        Parameters
        ----------
        query: str
            path for the file with the proteome, either GenBank or fasta.
        db: str
            Path to the BLAST database to be used.
        out: str
            Path to alignment results in tabular format.
        taxonlist: str
            Comma separated taxid values for positive filtering of the selected db.
        taxidlist: str
            Path to the file with the taxid values for positive filtering of the selected db.
        task: str
            BLASTP task name blasp, fast-blastp, etc ...
        matrix: str
            Substitution matrix.
        evalue: decimal
            Evalue filter.
        gapextend: int
            Penalty for gap extension.
        gapopen: int
            Penalty for gap opening.
        """
        self.engine_dict = coast_var.aln_options["blast"]
        self.merlin_engine = "BlastP"
        self.threads = ""
        self.tool_command = "blastp"

        self.query = os.path.basename(query)
        self.db = db
        self.taxidlist = taxidlist
        self.taxonlist = taxonlist
        self.task = task
        self.matrix = matrix
        self.evalue = evalue
        self.gapextend = gapextend
        self.gapopen = gapopen
        self.outfmt = f'"6 {coast_var.output_format["blast"]}"'
        self.out = out
        self.workspace = os.path.dirname(query)

    def generic_aligner(self, db="", query="", out=""):
        """
        Generates generic alignment command for BLAST and DIAMOND tools.
        if not provided special folders, it will use the workspace

        Parameters
        ----------
        db: str
            Path for the alignment database.
        query: str
            Path for the query file.
        out: str
            Path for the alignment result file.

        Returns
        -------
        str
            Alignment run command.
        """
        _db, _query, _out = self.db, self.query, self.out
        self.db, self.query, self.out = db, query, out

        for option in self.engine_dict.keys():
            if getattr(self, option):
                self.tool_command = " ".join([self.tool_command, f'{self.engine_dict[option]}{getattr(self, option)}'])

        self.db, self.query, self.out = _db, _query, _out
        return self.tool_command

    def run_local(self, threads):
        """
        Runs batch alignment locally.

        Parameters
        ----------
        threads: int
            Number of threads to run the alignment tool.
        """
        self.threads = threads
        subprocess.call(self.generic_aligner(db=self.db,
                                             query=os.path.join(self.workspace, self.query),
                                             out=os.path.join(self.workspace, self.out)),
                        shell=True)

    def container_call(self, aligner):
        """
        Creates container based query for BLAST alignment, by adding container run parameters to the base tool
        specific ones.

        Parameters
        ----------
        aligner: str, func
            Tool specific commands.

        Returns
        -------
        str
            Container run command.
        """
        # https://github.com/ncbi/blast_plus_docs
        cont_base = f'{self.containerization} run --rm ' \
                    '-v $HOME/blastdb:/blast/blastdb:ro ' \
                    '-v $HOME/blastdb_custom:/blast/blastdb_custom:ro ' \
                    f'-v {self.workspace}:/blast/queries:ro ' \
                    f'-v {self.workspace}:/blast/results:rw ' \
                    f'-w /blast {self.image}'
        return " ".join([cont_base, aligner])

    def run_container(self, containerization="docker", image=None, fdb="/blast/blastdb",
                      fquery="/blast/queries", fout="/blast/results"):
        """
        Runs batch alignment in a local container, this container is only for the alignment engine.

        Parameters
        ----------
        containerization
            Containerization engine to be used (docker, podman).
        image: str
            Alignment engine container image to be used.
        fdb: str
            Internal container folder for the alignment databases.
        fquery: str
            Internal container folder for the query file.
        fout: str
            Internal container folder for the alignment result file.
        """

        # Validate type of engine
        if containerization in ["docker", "podman"]:
            self.containerization = containerization
            if image:
                self.image = image
            else:
                self.image = coast_var.default_blast_container

            try:
                base_command = self.generic_aligner(db="/".join([fdb, os.path.basename(self.db)]),
                                                    query="/".join([fquery, self.query]),
                                                    out="/".join([fout, self.out])
                                                    )
                subprocess.call(self.container_call(base_command), shell=True)
            except:
                raise Exception("Could not run the necessary commands")
        else:
            raise Exception("Not a compatible containerization engine")

    def merlin_parameters(self):
        """
        Generates a merlin api call parameters.

        Returns
        -------
        str
            Parameters to be written to file.
        """
        query_parameters = []

        for option in ["db", "outfmt", "taxonlist", "taxidlist", "matrix", "evalue", "gapopen", "gapextend"]:
            if getattr(self, option):
                query_parameters.append(f'{self.engine_dict[option].strip("-").strip(" ")} = {os.path.basename(getattr(self, option))}')

        return query_parameters

    def request_merlin(self, files, address, email, raw_output="BlastResults.out", command="BlastP"):
        """
        Call api, fowards request files and confirms returns

        Returns
        -------
        email: str
            Email used for request logging.
        address: str
            Api server base address for request.
        file: list of str
            List of files to be used
        raw_output:
            Merlin given name to the output, to be changed to the coast tool defined one.
        """
        # TODO: check if in the proper network
        multiple_files = [('files', (os.path.basename(file), open(file, 'rb'))) for file in files]

        r = requests.post(f'{address}submitRestPlugin/{command}/{email}', files=multiple_files, timeout=2)
        if r.status_code == 201:
            status, timeout, interval = 201, 360000, 10  # in seconds
            attempts = timeout / interval
            attempt = 0
            message = json.loads(r.text)
            submissionID = message["submissionID"]
            while status in [201, 202] and attempt < attempts:
                time.sleep(10)
                attempt += 1
                status_response = requests.get(f'{address}/status/{email}/{submissionID}', timeout=3)
                status = status_response.status_code
            if status == 200:
                request.urlretrieve(f'{address}/download/{email}/{submissionID}',
                                    os.path.join(self.workspace, "merlin_run.zip"))
                unpack_archive(os.path.join(self.workspace, "merlin_run.zip"), extract_dir=self.workspace)
                result_diamond = hashlib.md5(open(os.path.join(self.workspace, raw_output), 'rb').read()).hexdigest()
                with open(os.path.join(self.workspace, 'checksum_BlastResults.md5')) as f:
                    first_line = f.readline().strip()
                if not result_diamond == first_line:
                    raise Exception("Output doesn't match md5 signature provided by the server.")
                os.rename(os.path.join(self.workspace, raw_output), os.path.join(self.workspace, self.out))
            else:
                raise Exception(f"{status_response.status_code}: {status_response.reason}")
        else:
            raise Exception(f"{r.status_code}: {r.reason}")

    def run_merlin(self, email, address='http://palsson.di.uminho.pt:6082/'):
        """
        Sends merlin api alignment request.

        Parameters
        ----------
        email: str
            Email used for request logging.
        address: str
            Api server base address for request.
        """
        protein_path = os.path.join(self.workspace, self.query)
        parameters_file_path = os.path.join(self.workspace, "SubmissionParameters.txt")
        files = [protein_path, parameters_file_path]
        # Generate params
        with open(parameters_file_path, mode='wt', encoding='utf-8') as parameters_file:
            parameters_file.write('\n'.join(self.merlin_parameters()))

        if self.taxidlist:
            files.append(self.taxidlist)
        self.request_merlin(files, address, email, raw_output="BlastResults.out")


class DiamondBatch(BlastBatch):
    def __init__(self, query, db, out="blast_results.tab", taxonlist="", taxidlist="", sensitivity="", matrix="",
                 evalue="", gapextend="", gapopen=""):
        """
        Formulate BLAST+ command line query.

        Parameters
        ----------
        query: str
            path for the file with the proteome, either GenBank or fasta.
        db: str
            Path to the BLAST database to be used.
        out: str
            Path to alignment results in tabular format.
        taxonlist: str
            Comma separated taxid values for positive filtering of the selected db.
        taxidlist: str
            Path to the file with the taxid values for positive filtering of the selected db.
        matrix: str
            Substitution matrix.
        evalue: decimal
            Evalue filter.
        gapextend: int
            Penalty for gap extension.
        gapopen: int
            Penalty for gap opening.
        """
        super().__init__(query, db, out, taxonlist, taxidlist, matrix, evalue, gapextend, gapopen)

        self.engine_dict = coast_var.aln_options["diamond"]
        self.merlin_engine = "DiamondP"

        self.sensitivity = sensitivity
        self.outfmt = f"6 {coast_var.output_format['diamond']}"

    def merlin_parameters(self):
        """
        Generates a merlin api call parameters.

        Returns
        -------
        str
            Parameters to be written to file.
        """
        query_parameters = []

        for option in ["db", "outfmt", "taxonlist", "matrix", "evalue", "gapopen", "gapextend"]:
            if getattr(self, option):
                query_parameters.append(f'{self.engine_dict[option].strip("--").strip(" ")} = {getattr(self, option)}')

        return query_parameters

    def run_merlin(self, email, address='http://palsson.di.uminho.pt:6082/'):
        """
        Sends merlin api alignment request.

        Parameters
        ----------
        email: str
            Email used for request logging.
        address: str
            Api server base address for request.
        """
        protein_path = os.path.join(self.workspace, self.query)
        parameters_file_path = os.path.join(self.workspace, "SubmissionParameters.txt")
        files = [protein_path, parameters_file_path]
        # Generate params
        with open(parameters_file_path, mode='wt', encoding='utf-8') as parameters_file:
            parameters_file.write('\n'.join(self.merlin_parameters()))

        self.request_merlin(files, address, email, raw_output="DiamondResults.out", command="DiamondP")
