import argparse
import os

from coast.sm import SimilarityMatrix
from coast.input import loader


def sm_cli():
    parser = argparse.ArgumentParser(description="coast - Comparative Omnics Alignment Search Tool v0.1")
    # TODO: add file validation
    parser.add_argument("query",
                        help="Produce the similarity matrices from a GenBank or FASTA file.",
                        type=str)
    parser.add_argument("results",
                        help="Path to the alignment results from the original coast run.",
                        type=str)
    parser.add_argument("-o", "--out",
                        help="The directory where the output will be delivered.",
                        type=str)
    parser.add_argument("-f", "--outfmt",
                        help="Choose the output format that will be used for the produced data. It can either be the "
                             "tabular, human readable text file or full workspace with intermediary files and extra "
                             "results.",
                        type=str,
                        nargs='*',
                        default=['b', 'a', 'r'])
    parser.add_argument("-k", "--keywords",
                        help="This option is only taken into account when a GenBank File is provided, in such cases, "
                             "it identifies the annotation keywords used to identify proteins, default is only using "
                             "CDS.",
                        type=str)
    parser.add_argument("--aai",
                        help="Minimum Average Amino Acid Identity, for reporting purposes",
                        default=10,
                        type=int)
    parser.add_argument("--id",
                        help="Minimum Amino Acid Identity, for hit selection for AAIbd calculation",
                        default=40,
                        type=int)
    parser.add_argument("--cov",
                        help="Minimum coverage, for hit selection for AAIbd calculation",
                        default=50,
                        type=int)
    parser.add_argument('--quiet', dest='verbose', action='store_false',
                        help="Deactivate verbose terminal output along the created files."
                             "This is already the default behaviour."
                        )
    parser.add_argument('--verbose', dest='verbose', action='store_true',
                        help="Activate verbose terminal output along the created files."
                        )
    parser.add_argument('--filter_hypothetical', dest='hypothetical', action='store_true',
                        help="Filters proteins with 'hypothetical' in their description, use with caution."
                        )
    parser.set_defaults(verbose=False, out=os.getcwd(), engine="blast", keywords="CDS", hypothetical=False)

    args = parser.parse_args()
    for arg in args.outfmt:
        if arg not in ["a", "b", "r"]:
            parser.error(f"The output format '{arg}' is not in the authorized list b/a/r.")
    return args


def coast_report():
    cli = sm_cli()
    job = SimilarityMatrix()
    sequences = loader(cli.query, cli.keywords, hypothetical=cli.hypothetical)
    job.from_results(sub_sequences=sequences, results_path=cli.results, min_cov=cli.cov, min_id=cli.id)

    if "b" in cli.outfmt:
        job.tabular_sm[(job.tabular_sm['pident'] != 0)].to_csv(path_or_buf=os.path.join(cli.out, "bh_results.tab"),
                                                               sep="\t", index=False)
    if "a" in cli.outfmt:
        job.tabular_agr_sm.to_csv(path_or_buf=os.path.join(cli.out, "coast_results.tab"), sep="\t", index=False)
    if "r" in cli.outfmt:
        job.html_agr_sm(output_path=os.path.join(cli.out, "coast_report.html"), min_aai=cli.aai)
    if cli.verbose:
        job.print_agr_sm(min_aai=cli.aai)


# Run CLI
if __name__ == '__main__':
    coast_report()
