import argparse
import os
import datetime

from coast.coast import COAST


def coast_cli():
    parser = argparse.ArgumentParser(description="coast - Comparative Omnics Alignment Search Tool v0.1")
    # TODO: add file validation
    parser.add_argument("query",
                        help="Produce the similarity matrices from a GenBank or FASTA file.",
                        type=str)
    parser.add_argument("database",
                        help="Directory for the database against which the query will be run.",
                        type=str)
    parser.add_argument("-o", "--out",
                        help="The directory where the output will be delivered.",
                        type=str)
    parser.add_argument("-e", "--engine",
                        choices=['blast', 'diamond'],
                        help="Alignment Engine.",
                        type=str)
    parser.add_argument("-f", "--outfmt",
                        help="Choose the output format that will be used for the produced data. It can either be the "
                             "tabular, human readable text file or full workspace with intermediary files and extra "
                             "results.",
                        type=str,
                        nargs='*',
                        default=['b', 'a', 'r'])
    parser.add_argument("-k", "--keywords",
                        help="This option is only taken into account when a GenBank File is provided, in such cases, "
                             "it identifies the annotation keywords used to identify proteins, default is only using "
                             "CDS.",
                        type=str)
    taxonomy = parser.add_mutually_exclusive_group()
    taxonomy.add_argument("--taxidlist",
                          help="File with NCBI taxonomic IDs to filter the database by.",
                          type=str)
    taxonomy.add_argument("--taxonlist",
                          help="Comma-separated list of NCBI taxonomic IDs to filter the database by.",
                          type=str)
    engine_options = parser.add_mutually_exclusive_group()
    engine_options.add_argument("--task",
                                choices=['blastp', 'blastp-fast', 'blastp-short'],
                                help="Setting for the blast task. Read the BLAST documentation for more information.",
                                type=str)
    engine_options.add_argument("--sens",
                                choices=['mid-sensitive', 'sensitive', 'more-sensitive', 'very-sensitive',
                                         'ultra-sensitive'],
                                help="Setting for the diamond alignment tool sensibility.",
                                type=str)
    parser.add_argument("--matrix",
                        choices=['BLOSUM45', 'BLOSUM50', 'BLOSUM62', 'BLOSUM80', 'BLOSUM90', 'PAM250', 'PAM70',
                                 'PAM30'],
                        help="Scoring matrix",
                        type=str)
    parser.add_argument("--evalue",
                        help="E-Value Threshold",
                        type=str)
    parser.add_argument("--gapopen",
                        help="Gap open penalty.",
                        type=str)
    parser.add_argument("--gapextend",
                        help="Gap extension penalty.",
                        type=str)
    parser.add_argument("--threads",
                        help="Number of threads available for the engine to run. Read the respective engine's "
                             "documentation for further details.",
                        type=str)
    parser.add_argument("-cont", "--containerization",
                        choices=['docker', 'podman'],
                        help="Containerization system, if not provided will run with local software. It must be OCI "
                             "complaint. Be sure that your user has docker execution "
                             "priviliges, so you dont need to provide sudo authorization.",
                        type=str)
    parser.add_argument("-img", "--image",
                        help="Image for the container to run the selected engine. Be sure that the container contains "
                             "the actual software installed. If you dont provide a -cont argument, it will be run "
                             "with docker as default. If your system lacks docker it will error out",
                        type=str)
    parser.add_argument("-w", "--web",
                        help="Merlin server address.",
                        default="http://palsson.di.uminho.pt:6082/",
                        type=str)
    parser.add_argument("--email",
                        help="Email for the Merlin server run.",
                        type=str)
    parser.add_argument("--aai",
                        help="Minimum Average Amino Acid Identity, for reporting purposes",
                        default=5,
                        type=int)
    parser.add_argument("--id",
                        help="Minimum Amino Acid Identity, for hit selection for AAIbd calculation",
                        default=40,
                        type=int)
    parser.add_argument("--cov",
                        help="Minimum coverage, for hit selection for AAIbd calculation",
                        default=50,
                        type=int)
    parser.add_argument('--quiet', dest='verbose', action='store_false',
                        help="Deactivate verbose terminal output along the created files."
                             "This is already the default behaviour."
                        )
    parser.add_argument('--verbose', dest='verbose', action='store_true',
                        help="Activate verbose terminal output along the created files."
                        )
    parser.add_argument('--generate_workspace', dest='gen_ws', action='store_true',
                        help="Create the workspace inside a named and dated directory."
                        )
    parser.add_argument('--inplace_workspace', dest='gen_ws', action='store_false',
                        help="Generate all files straight to the current working directory or the specified directory."
                        )
    parser.set_defaults(verbose=False, out=os.getcwd(), engine="blast", keywords="CDS", gen_ws=True)
    args = parser.parse_args()

    # defaults
    if args.threads and (args.image or args.containerization):
        parser.error("Cannot run in local and container context.")
    if args.threads and (args.web or args.email):
        parser.error("Cannot run in local and web context.")
    if (args.web or args.email) and (args.image or args.containerization):
        parser.error("Cannot run in web and container context.")
    if args.task and args.engine != "blast":
        parser.error("The blast task options are exclusive to the blast mode.")
    if args.sens and args.engine != "diamond":
        parser.error("The diamond sensitivity options are exclusive to the diamond mode.")
    if args.web and not args.email:
        parser.error("To run the Merlin web mode an email is required.")

    for arg in args.outfmt:
        if arg not in ["a", "b", "r"]:
            parser.error(f"The output format '{arg}' is not in the authorized list b/a/r.")
    return args


def coast_search():
    cli = coast_cli()
    job = COAST()
    start_time = datetime.datetime.now()
    workspace = job.generate_workspace(input_file=cli.query,
                                       directory=cli.out,
                                       key=cli.keywords,
                                       generate_workspace=cli.gen_ws)
    log = open(os.path.join(workspace, "log.txt"), 'w')
    log.write(f"start workspace:{start_time}\n")
    log.write(f"start alignment:{datetime.datetime.now()}\n")
    if cli.engine == "blast":
        job.blast_aln(db=cli.database,
                      taxonlist=cli.taxonlist,
                      taxidlist=cli.taxidlist,
                      task=cli.task,
                      matrix=cli.matrix,
                      evalue=cli.evalue,
                      gapextend=cli.gapextend,
                      gapopen=cli.gapopen)
    elif cli.engine == "diamond":
        job.diamond_aln(db=cli.database,
                        taxonlist=cli.taxonlist,
                        taxidlist=cli.taxidlist,
                        sensitivity=cli.sensitivity,
                        matrix=cli.matrix,
                        evalue=cli.evalue,
                        gapextend=cli.gapextend,
                        gapopen=cli.gapopen)

    if cli.containerization or cli.image:
        job.aln.run_container(containerization=cli.containerization, image=cli.image)
    elif cli.web:
        job.aln.run_merlin(email=cli.email, address=cli.web)
    else:
        job.aln.run_local(threads=cli.threads)

    log.write(f"start identification:{datetime.datetime.now()}\n")
    sm = job.cumulative_identification(min_cov=cli.cov, min_id=cli.id)

    if "b" in cli.outfmt:
        job.best_hits()
    if "a" in cli.outfmt:
        job.aggregated_results()
    if "r" in cli.outfmt:
        job.generate_report(min_aai=cli.aai)
    log.write(f"end work:{datetime.datetime.now()}\n")
    if cli.verbose:
        sm.print_agr_sm(min_aai=cli.aai)


# Run CLI
if __name__ == '__main__':
    coast_search()
