import os
from Bio import SeqIO
import datetime

from coast.sm import SimilarityMatrix
from coast.input import loader
from coast.aln import BlastBatch, DiamondBatch


class COAST:
    """
    Interface object for a whole run of the tool.
    """

    def __init__(self):
        self.input = ""
        self.sequences = {}
        self.sm = None
        self.workspace = ""
        self.aln = None
        self.organism = ""

    def load_input(self, file, key="CDS"):
        """
        Loading the input file for the protome.

        Parameters
        ----------
        file: str
            path for the file with the proteome, either GenBank or fasta.
        key: str
            Keyword that represents the feature type to identify the coding sequences in GenBank files.

        Returns
        -------
        sub_sequences: dict of Bio.SeqRecord
            Sequences of the reference proteome.

        """
        self.sequences = loader(file, key)
        SeqIO.write(self.sequences.values(), os.path.join(self.workspace, "protein.faa"), "fasta")
        return self.sequences

    def generate_workspace(self, input_file, directory="", key="CDS", generate_workspace=True):
        """
        Generates a new folder where all the analysis and temp files will be saved. This folder will be located at
        either the users cwd, as default or other of the user's choosing. The user is also provided the option for
        automatic clean up after the execution, to look for this function look into the

        naming convention: "input file name"_COAST_"datetime in seconds"

        Parameters
        ----------
        generate_workspace: bool
            Indicate if tool should create the workspace inside a new folder or straight into the cwd.
        input_file: str
            Path for the file with the proteome, either GenBank or fasta.
        directory: str
            Path for the desired location for the workspace, if not that of the input.
        key: str
            Keyword that represents the feature type to identify the coding sequences in GenBank files.

        Returns
        -------
        str
            Workspace path.

        """
        # finds the name of the loaded file, that will be used as the base to build the rest of the dir name
        self.input = input_file
        base_name = os.path.basename(self.input).rsplit(".", 1)[0]

        if not dir:
            directory = os.path.dirname(self.input)

        if generate_workspace:
            dir_name = base_name + "_COAST_" + datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
            self.workspace = os.path.join(directory, dir_name)
            os.mkdir(self.workspace)
        else:
            self.workspace = directory

        self.organism = base_name
        self.load_input(input_file, key)
        return self.workspace

    def blast_aln(self,
                  db,
                  taxonlist="",
                  taxidlist="",
                  task="",
                  matrix="",
                  evalue="",
                  gapextend="",
                  gapopen=""):
        """
        Formulate BLAST+ command line query.

        Parameters
        ----------
        db: str
            Path to the BLAST database to be used.
        taxonlist: str
            Comma separated taxid values for positive filtering of the selected db.
        taxidlist: str
            Path to the file with the taxid values for positive filtering of the selected db.
        task: str
            BLASTP task name blasp, fast-blastp, etc ...
        matrix: str
            Substitution matrix.
        evalue: decimal
            Evalue filter.
        gapextend: int
            Penalty for gap extension.
        gapopen: int
            Penalty for gap opening.

        Returns
        -------
        str
            Query.
        """
        if not self.aln:
            try:
                self.aln = BlastBatch(query=os.path.join(self.workspace, "protein.faa"),
                                      out="blast_results.tab",
                                      db=db,
                                      taxonlist=taxonlist,
                                      taxidlist=taxidlist,
                                      task=task,
                                      matrix=matrix,
                                      evalue=evalue,
                                      gapextend=gapextend,
                                      gapopen=gapopen)
                return self.aln
            except:
                raise Exception("Could not produce alignment.")
        else:
            raise Exception("Alignment has been run.")

    def diamond_aln(self,
                    db,
                    taxonlist="",
                    taxidlist="",
                    sensitivity="",
                    matrix="",
                    evalue="",
                    gapextend="",
                    gapopen=""):
        """
        Formulate DIAMOND command line query.

        Parameters
        ----------
        db: str
            Path to the BLAST database to be used.
        taxonlist: str
            Comma separated taxid values for positive filtering of the selected db.
        taxidlist: str
            Path to the file with the taxid values for positive filtering of the selected db.
        sensitivity: str
            Diammond sensitivity setting.
        matrix: str
            Substitution matrix.
        evalue: decimal
            Evalue filter.
        gapextend: int
            Penalty for gap extension.
        gapopen: int
            Penalty for gap opening.

        Returns
        -------
        str
            Query.
        """
        if not self.aln:
            try:
                self.aln = DiamondBatch(query=os.path.join(self.workspace, "protein.faa"),
                                        out="blast_results.tab",
                                        db=db,
                                        taxonlist=taxonlist,
                                        taxidlist=taxidlist,
                                        sensitivity=sensitivity,
                                        matrix=matrix,
                                        evalue=evalue,
                                        gapextend=gapextend,
                                        gapopen=gapopen)
                return self.aln
            except:
                raise Exception("Could not produce alignment.")
        else:
            raise Exception("Alignment has been run.")

    def cumulative_identification(self, results_path="", min_cov=50, min_id=40):
        """
        Intermediate step necessary for the production of a similarity matrix, this method will load the batch
        alignment results and produce a similarity matrix that might be used for further output generation.

        Parameters
        ----------
        results_path: str
            Path to alignment results in tabular format.
        min_id: int
            Minimum Amino Acid Identity, for hit selection for AAIbd calculation.
        min_cov: int
            Minimum coverage, for hit selection for AAIbd calculation.

        Returns
        -------
        coast.SimilarityMatrix
            Similarity matrix for the current job
        """
        if not results_path:
            results_path = os.path.join(self.workspace, "blast_results.tab")
        if self.aln or os.path.isdir(results_path):
            # try:
            self.sm = SimilarityMatrix().from_results(sub_sequences=self.sequences, results_path=results_path,
                                                      organism_id=self.organism, min_id=min_id, min_cov=min_cov)
            return self.sm
            # except:
            #     raise Exception("It was not possible to load the necessary files, make sure they exist.")
        else:
            raise Exception("No valid alignment results were found, make sure that you have run the alignment step or "
                            "that you have a valid file.")

    def best_hits(self):
        """ Shortcut to the save as csv pandas function for the tabular similarity matrix.
        """
        if not self.sm:
            self.cumulative_identification()
        self.sm.tabular_sm[(self.sm.tabular_sm['pident'] != 0)].to_csv(path_or_buf=os.path.join(self.workspace, "bh_results.tab"), sep="\t", index=False)

    def aggregated_results(self):
        """ Shortcut to the save as csv pandas function for the tabular aggregated similarity matrix.
        """
        if not self.sm:
            self.cumulative_identification()
        self.sm.tabular_agr_sm.to_csv(path_or_buf=os.path.join(self.workspace, "coast_results.tab"), sep="\t", index=False)

    def generate_report(self, min_aai=10):
        """
        Shortcut to the the markdown report generation function for the SimilarityMatrix object, passing the
        filtering minimum aai score.

        Parameters
        ----------
        min_aai: int
             minimum aai score for filtering in the report
        """
        if not self.sm:
            self.cumulative_identification()
        self.sm.html_agr_sm(output_path=os.path.join(self.workspace, "coast_report.html"), min_aai=min_aai)
