from coast.input import dataframe_from_sequences
from coast.blast_results import BlastResults
import pandas as pd
import os
import matplotlib.pyplot as plt
from tabulate import tabulate
import seaborn as sns
import base64
import dominate
from dominate.tags import *
from dominate.util import raw
import datetime
import numpy as np


class SimilarityMatrix:
    def __init__(self):
        self.min_cov = None
        self.min_id = None
        self.circos = None
        self.tabular_sm = None
        self.tabular_sm_cov = None
        self.tabular_agr_sm = None
        self.aln = None
        self.bh_cov = None
        self.bh = None
        self.query = None
        self.organism = None
        self.compared_organism = None
        self.aai_heatmap = None

    def from_results(self, sub_sequences, results_path, organism_id="Unknown", min_cov=50, min_id=40):
        """ Generates a SimilarityMatrix object from a alignment file.
        Parameters
        ----------
        min_id: int
            Minimum Amino Acid Identity, for hit selection for AAIbd calculation.
        min_cov: int
            Minimum coverage, for hit selection for AAIbd calculation.
        sub_sequences: dict of Bio.SeqRecord
            Sequences of the reference proteome.
        results_path: str
            Path to alignment results in tabular format.
        organism_id: str
            Identifier of the reference proteome to be used for reporting purposes.
        """
        self.aln = BlastResults().from_file(results_path)
        self.bh = self.aln.best_hits_from_blast()
        self.query = dataframe_from_sequences(sub_sequences)
        self.organism = organism_id

        # auxiliary
        self.compared_organism = self.bh.get_organisms()
        self.aggregated_metrics(min_cov=min_cov)
        self.min_cov = min_cov
        self.min_id = min_id
        return self

    def cumulative_identification(self, min_cov=0):
        """
        Through a left join, it identifies for each taxid in the bh dataframe, and adds, the proteins that are missing
        from the query.

        Returns
        -------
        pandas.Dataframe
            Similarity matrix in a tabular format
        """
        try:
            targets = self.query.merge(pd.DataFrame(data=self.compared_organism, columns=["staxids"]), how="cross")
            self.tabular_sm = targets.merge(self.bh.results, how="left", on=["qseqid", "staxids"]).sort_values(
                by=['qseqid', 'pident'])
            if min_cov != 0:
                self.bh_cov = self.aln.best_hits_from_blast(min_cov=min_cov)
                self.tabular_sm_cov = self.bh.results
            return self.tabular_sm
        except:
            raise Exception(pd.__version__)

    def aggregated_metrics(self, min_cov=0):
        """
        Calculate the aggregated metrics (aai, hitspp, etc) for the similarity matrix, aggregated by organism (staxids).

        Returns
        -------
        pandas.Dataframe
            Aggregated metrics for the similarity matrix
        """
        if not isinstance(self.tabular_sm, pd.DataFrame):
            # To check dataframe the normal method is not compatible
            self.cumulative_identification(min_cov=min_cov)
        if isinstance(self.tabular_sm, pd.DataFrame):
            if self.tabular_sm.empty:
                self.cumulative_identification(min_cov=min_cov)

        # Initiate aggregated df and calculates unweighted AAIc
        self.tabular_sm.pident = self.tabular_sm.pident.fillna(0)
        self.tabular_agr_sm = self.tabular_sm.groupby(by=["staxids", "sscinames", "sskingdoms"], as_index=False).agg(
            aaic=("pident", "sum"))

        total_query_sequence = len(self.query)
        self.tabular_agr_sm.aaic = self.tabular_agr_sm.aaic / total_query_sequence
        self.tabular_agr_sm.aaic = round(self.tabular_agr_sm.aaic, ndigits=3)

        # AAI bare-bones
        # _aaibd = self.tabular_sm.copy()
        # _aaibd.pident = _aaibd["pident"].replace(0, np.nan)
        # aaibd = _aaibd.groupby(by=["staxids"], as_index=False).agg(aaibd=("pident", "mean"))
        # self.tabular_agr_sm = self.tabular_agr_sm.merge(aaibd, how='left', on=["staxids"])
        # self.tabular_agr_sm["aaibd"] = round(self.tabular_agr_sm["aaibd"], ndigits=3)

        # AAIb/AAId
        if min_cov != 0:
            _aaibd_fil = self.tabular_sm_cov[self.tabular_sm_cov.pident > 40].copy()
        else:
            _aaibd_fil = self.tabular_sm[self.tabular_sm.pident > 40].copy()
        _aaibd_fil.pident = _aaibd_fil.pident.replace(0, np.nan)
        aaibd_fil = _aaibd_fil.groupby(by=["staxids"], as_index=False).agg(aaibd_fil=("pident", "mean"))
        self.tabular_agr_sm = self.tabular_agr_sm.merge(aaibd_fil, how='left', on=["staxids"])
        self.tabular_agr_sm["aaibd_fil"] = round(self.tabular_agr_sm["aaibd_fil"], ndigits=3)

        # Hits per protein
        count = self.aln.count_hits().groupby(by=["staxids"], as_index=False).agg(hitspp=("size", "sum"))
        self.tabular_agr_sm = self.tabular_agr_sm.merge(count, how='left', on=["staxids"])
        self.tabular_agr_sm["hitspp"] = self.tabular_agr_sm["hitspp"] / total_query_sequence
        self.tabular_agr_sm["hitspp"] = round(self.tabular_agr_sm["hitspp"], ndigits=3)

        # nº of match
        count = self.tabular_sm.groupby(by=["staxids"], as_index=False).agg(match=("nident", "count"))
        self.tabular_agr_sm = self.tabular_agr_sm.merge(count, how='left', on=["staxids"]).sort_values(by=['aaic'],
                                                                                                       ascending=[0])
        self.tabular_agr_sm["pmatch"] = self.tabular_agr_sm["match"] / total_query_sequence
        return self.tabular_agr_sm

    def filtered_agr_sm(self, min_aai):
        """
        Filtered version of the agregated metrics table, filtered by minimum AAI score.

        Parameters
        ----------
        min_aai: int
             Minimum aai score for filtering in the report.

        Returns
        -------
        pandas.Dataframe
            Aggregated metrics for the similarity matrix.
        """
        return self.tabular_agr_sm[(self.tabular_agr_sm['aaic'] >= min_aai)].sort_values(by=['aaic', 'hitspp'],
                                                                                         ascending=False)

    def draw_aai_heatmap(self, path, min_aai=10):
        df = self.tabular_sm.set_index(['staxids', 'qseqid'])['pident'].unstack().reset_index()
        df = df.merge(self.filtered_agr_sm(min_aai=min_aai), on=["staxids"])
        df = df.set_index('staxids').sort_values(by=['aaic'], ascending=[0])
        height = len(df.aaic)
        width = len(self.query['qseqid'])
        ratio = round(height / width, ndigits=1)

        df = df.drop(labels=["sscinames", "sskingdoms", "aaic", "aaibd", "aaibd_fil", "hitspp", "match"], axis=1)
        sns.set_theme(style="white")
        f, ax = plt.subplots(figsize=(4, 4 + ratio * 2))
        sns.heatmap(df, vmin=0, vmax=100, cmap="YlGnBu", square=False, cbar_kws={"shrink": .5})
        self.aai_heatmap = os.path.join(path, "heatmap_plot.png")
        plt.subplots_adjust(left=0, top=1, bottom=0)
        plt.savefig(self.aai_heatmap)
        return self.aai_heatmap

    def markdown_agr_sm(self, output_path, min_aai=10):
        """
        Produces a markdown report of the results.

        Parameters
        ----------
        output_path: str
            Path, including file name for the report to be saved at.

        min_aai: int
            Minimum aai score for filtering in the report.

        Returns
        ----------
        bool
            Execution confirmation
        """
        output_report = open(output_path, 'w')
        output_report.write("# coast Results" + "\n" + "\n")
        output_report.write("## Summary" + "\n" + "\n")
        output_report.write(self.filtered_agr_sm(min_aai).to_markdown(index=False, floatfmt=".3f") + "\n" + "\n")
        output_report.write("## AAI Heatmap and Protein Association" + "\n" + "\n")
        if not self.aai_heatmap:
            self.draw_aai_heatmap(path=os.path.dirname(output_path), min_aai=min_aai)
        output_report.write(f'![heatmap]({os.path.basename(self.aai_heatmap)})')
        output_report.close()
        return True

    def html_agr_sm(self, output_path, min_aai=10):
        """
        Produces a markdown report of the results.

        Parameters
        ----------
        output_path: str
            Path, including file name for the report to be saved at.

        min_aai: int
            Minimum aai score for filtering in the report.

        Returns
        ----------
        bool
            Execution confirmation
        """
        doc = dominate.document(title='COAST Report')
        if not self.aai_heatmap:
            self.draw_aai_heatmap(path=os.path.dirname(output_path), min_aai=min_aai)
        with open(self.aai_heatmap, "rb") as img_file:
            heatmap = base64.b64encode(img_file.read()).decode('utf-8')
        with doc:
            with div(id='header'):
                h1("COAST Report")
            with div():
                attr(cls='body')
                with div():
                    h2('Query Summary')
                    p(f"Organism: {self.organism}")
                    p(f"Number of proteins: {len(self.query)}")
                    h3(f"AAIc settings:")
                    p(f"AAIc filter: {min_aai}%")
                    h3(f"AAIbd settings:")
                    p(f"Minimum ID: {self.min_id}%")
                    p(f"Minimum Coverage: {self.min_cov}%")
                    p(f"Report Date: {datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}")
                with div():
                    h2('Table Summary')
                    td(raw(self.filtered_agr_sm(min_aai).to_html(index=False, float_format="%.3f")))
                with div():
                    h2('Amino Acid Identity Heatmap')
                    img('Heatmap', src="data:image/jpeg;base64," + heatmap)

        output_report = open(output_path, 'w')
        output_report.write(doc.render())
        output_report.close()
        return True

    def print_agr_sm(self, min_aai=10):
        """
        Prints command line tabulated summary of the aggregated metrics.

        Parameters
        ----------
        min_aai: int
            Minimum aai score for filtering in the report.
        """
        print(tabulate(self.filtered_agr_sm(min_aai), showindex=False, headers='keys', tablefmt='psql', floatfmt=".3f"))
