from coast.coast_var import output_format
import pandas as pd


class BlastResults:
    """
        This is a parser class for BLAST tabular 6 output formatted tables.
    """

    def __init__(self):
        self.results = None

    def explode_staxids(self):
        """ Breaks up hits with more than one species (taxids)
        """
        self.results['staxids'] = [str(x).split(';') for x in self.results['staxids']]
        self.results['sscinames'] = [str(x).split(';') for x in self.results['sscinames']]
        self.results = self.results.apply(pd.Series.explode)

    def from_file(self, file_path, header=""):
        """
        Loads tabular output from file

        Parameters
        ----------
        file_path: str
            path for the file, formatted in tabular 6 output, according to the column specification
        header: list of str
            column names for the tabular output, by order

        Returns
        -------
        COAST.BlastResults
            BlastResults object with the tabular data

        """
        if not header:
            header = output_format["blast"].split()
        self.results = pd.read_table(file_path, names=header)
        self.explode_staxids()
        return self

    def from_df(self, df):
        """
        Loads tabular -6 BLAST output from a pandas dataframe

        Parameters
        ----------
        pandas.Dataframe
            pandas dataframe loaded with tabular -6 BLAST outputs and WITH the header present

        Returns
        -------
        COAST.BlastResults
            References self BlastResults object with the tabular data

        """
        self.results = df
        self.explode_staxids()
        return self

    def best_hits_from_blast(self, min_cov=0):
        """
        Filters the best hits in a blast id, aggregated by query sequence and taxid, designed to simplify similarity
        matrix intake.

        Returns
        -------
        COAST.BlastResults
            References self BlastResults object with the tabular data

        """
        _results = self.results
        if min_cov > 0:
            _results = self.results[self.results.qcovs > min_cov]
        results = _results.sort_values('pident', ascending=False).drop_duplicates(['qseqid', 'staxids'])
        return BlastResults().from_df(results)

    def count_hits(self):
        """
        Lists the hits per combination of qseqid and staxids.

        Returns
        -------
        pandas.DataFrame
            Columns:
                qseqid: str
                staxids: str
                hits: int

        """
        return self.results.groupby(by=["qseqid", "staxids"], as_index=False).size()

    def get_organisms(self):
        """
        For a given tabular output, list all the unique organism (staxids).

        Returns
        -------
        list of str
            staxids for each unique organism

        """
        return self.results.staxids.unique()
